// ================================================================= //
// (C) The Linbox Group 1999
// Time-stamp: <05 Apr 00 16:10:12 Jean-Guillaume.Dumas@imag.fr> 
// ================================================================= //

#ifndef _GIV_POLY1_P_ADIC_H_
#define _GIV_POLY1_P_ADIC_H_
#include <givinteger.h>
#include <givpoly1.h>

template<class Domain, class Tag> class Poly1PadicDom;



template<class Domain>
class Poly1PadicDom<Domain,Dense> : public Poly1Dom<Domain,Dense>, public IntegerDom {
    typedef typename Poly1Dom<Domain,Dense>::Rep      Rep;
public:
    typedef typename Poly1Dom<Domain,Dense>::element  element;
    typedef typename Poly1Dom<Domain,Dense>::element  pol_element;
    typedef typename IntegerDom::element              int_element;

    Poly1PadicDom (Domain& d, const Indeter& X) : Poly1Dom<Domain,Dense> (d,X), IntegerDom() {}
    Poly1PadicDom (const Poly1Dom<Domain,Dense>& P) : Poly1Dom<Domain,Dense> (P), IntegerDom() {}
    Poly1PadicDom (const Poly1Dom<Domain,Dense>& P, const IntegerDom& D) : Poly1Dom<Domain,Dense> (P), IntegerDom(D) {}


        // Horner like evaluation of the polynomial for p = _domain.size()
    IntegerDom::element& eval( IntegerDom::element& E, const element& P) {
        typename element::const_reverse_iterator pi = P.rbegin();
        IntegerDom::init(E, _domain.write(*pi) );
        for (++pi;pi != P.rend();++pi) {
            IntegerDom::mulin(E, _domain.size() );
            IntegerDom::addin(E, _domain.write(*pi) );
        }
        return E;
    }

        // Gain is 40% to 75% compared to Integers !!!
    unsigned long& eval( unsigned long& E, const element& P) {
        typename element::const_reverse_iterator pi = P.rbegin();
        _domain.write(E,*pi);
        for (++pi;pi != P.rend();++pi) {
            E *= _domain.size();
            E += _domain.write(*pi);
        }
        return E;
    }


        // Reconstruction of the radix decomposition mod _domain.size()
        // E into a polynomial P of degree n-1 or of lowest degree if n==0.
    element& radix(element& P, const IntegerDom::element& E, long n = 0) {
        if (n < 1) n = logp(E,_domain.size()) + 1;
        if (n == 1) {
            typename Domain::element e;
            return Poly1Dom<Domain,Dense>::init(P, Degree(0), _domain.init(e, E) );
        }
        IntegerDom::element iq, ir;
        element Q; 
            long t = (n+1)/2;
            IntegerDom::element q;
            IntegerDom::pow(q, _domain.size(), t);
            IntegerDom::divmod(iq, ir, E, q);
            radix(Q, iq, n-t);
            radix(P, ir, t);
            Degree dp; degree(dp,P); ++dp;
            for(long i=t; dp<i; --i)
                P.push_back(_domain.zero);
        P.insert(P.end(),Q.begin(),Q.end());
        return setdegree(P);
    } 
};

#endif 
