// ==========================================================================
// file: givgfq.inl
// Description:
//   Arithmetic on GF(q)
// Bugs:
// Time-stamp: <01 Sep 00 14:03:45 Jean-Guillaume.Dumas@imag.fr> 
// ==========================================================================

// Warning : valid iff b != c
#ifndef __GIVARO_COUNT__

#define _GIVARO_GFQ_ADD(c, a, b, mun, plun) { if ((b)==0) (c)=(a); else if ((a)==0) (c)=(b); else { \
                                   (c) = (a)-(b); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }

#define _GIVARO_GFQ_NEG(res, a, mo, mun) { if ( (a)==0 ) (res)=0; else\
                              { (res) = (a) - (mo) ; (res) = ((res)>0)?(res):(res)+(mun); } }

// Warning : valid iff a != c
// if not use AUTOSUB ...
#define _GIVARO_GFQ_SUB(c, a, b, mo, mun, plun) { if ((a)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)==0) (c)=(a); else { \
                                   (c) = (b)-(a)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(a); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }
#define _GIVARO_GFQ_AUTOSUB(c, b, mo, mun, plun) { if ((c)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)!=0) { \
                                   (c) = (c)-(b)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c)-(mo):(c)+(mo); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }



// #define _GIVARO_GFQ_MUL(res, a, b, mun) { if ( ((a)==0) || ((b)==0) ) { (res) =0; } else { (res) = (((res) = (a)+(b)-(mun) )>0)?(res):(res)+ (mun); } }
 #define _GIVARO_GFQ_MUL(res, a, b, mun) { if ( ((a)==0) || ((b)==0) ) { (res) =0; } else { (res) = (((res) = (a)+(b) )>(mun))?(res)-(mun):(res); } }

// JGD 02.04.1998 :  if a==1, a /= a used to be --> 0 !!!
#define _GIVARO_GFQ_INV(res, a, mun)    { (res) = (mun)-(a); (res)=(res)?(res):(mun); }

#define _GIVARO_GFQ_DIV(res, a, b, mun) {  \
if ( (a)==0 ) { (res)=0; } else { (res) = (((res)=(a)-(b))>0)?(res):(res)+(mun); } }



#define _GIVARO_GFQ_SQ(res, a, mun) { if ( (a)==0) (res) = 0; else \
                                  { (res) = ( (a) << 1) - (mun); \
                                    (res) = ((res)>0)?(res):(res)+ (mun); } }

// plun -> 1+^c - (q-1) !!!
// Warning : valid iff b != c
#define _GIVARO_GFQ_SQADD(c,a,b,mun,plun) { \
if ((a)==0) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((  (c)=((a) << 1) - (mun)        )>0)?(c):(c) + (mun); \
} else { \
       (c) = ((    (c) = ((a) << 1)-(b)-(mun)             )<0)?(c)+(mun):(c); \
       if (  (c) = (plun)[(UTT)(((c)>0)?(c):(c)+(mun))]     ) { \
                (c) = ((    (c) = (c)+(b)         )>0)?(c):(c)+(mun); } \
     }\
}

// Warning : valid iff b != c
#define _GIVARO_GFQ_MULADD(c,a1,a2,b,mun,plun) { \
if (((a1)==0) || ((a2)==0)) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((    (c)=(a1)+(a2) - (mun)       )>0)?(c):(c) + (mun); \
} else { \
       (c) = ((    (c) = (a1)+(a2)-(b)-(mun)        )<0)?(c)+(mun):(c); \
       if ( (c) = (plun)[(UTT)( ((c)>0)?(c):(c)+(mun)   )]  ) { \
                (c) = ((    (c) = (c)+(b)        )>0)?(c):(c)+(mun); }\
     }\
}

// Warning : valid iff b != c
#define _GIVARO_GFQ_MULSUB(c,a1,a2,b,mo,mun,plun) { \
if (((a1)==0) || ((a2)==0)) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((    (c)=(a1)+(a2) - (mo) -(mun)       )>0)?(c):(c) + (mun); \
       (c) = (c)>0?(c):(c) + (mun); \
} else { \
       (c) = ((    (c) = (a1)+(a2)-(b)-(mun) - (mo)       )<0)?(c)+(mun):(c); \
       (c) = (c)<0?(c)+(mun):(c); \
       if ( (c) = (plun)[(UTT)( ((c)>0)?(c):(c)+(mun)   )]  ) { \
                (c) = ((    (c) = (c)+(b)        )>0)?(c):(c)+(mun); }\
     }\
}



#else


// Warning : valid iff b != c

#define _GIVARO_GFQ_ADD(c, a, b, mun, plun) { ++_add_call; if ((b)==0) (c)=(a); else if ((a)==0) (c)=(b); else { \
                                   (c) = (a)-(b); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } ++_add_count; } }

#define _GIVARO_GFQ_NEG(res, a, mo, mun) { ++_neg_call; if ( (a)==0 ) (res)=0; else\
                              { (res) = (a) - (mo) ; (res) = ((res)>0)?(res):(res)+(mun); ++_neg_count; } }

// Warning : valid iff a != c
// if not use AUTOSUB ...
#define _GIVARO_GFQ_SUB(c, a, b, mo, mun, plun) { ++_sub_call; if ((a)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)==0) (c)=(a); else { \
                                   (c) = (b)-(a)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(a); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } ++_sub_count; } }
#define _GIVARO_GFQ_AUTOSUB(c, b, mo, mun, plun) { ++_sub_call; if ((c)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)!=0) { \
                                   (c) = (c)-(b)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c)-(mo):(c)+(mo); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } ++_sub_count; } }


#define _GIVARO_GFQ_MUL(res, a, b, mun) { ++_mul_call; if ( ((a)==0) || ((b)==0) ) { (res) =0; } else { (res) = (((res) = (a)+(b)-(mun) )>0)?(res):(res)+ (mun); ++_mul_count; } }

// JGD 02.04.1998 :  if a==1, a /= a used to be --> 0 !!!
#define _GIVARO_GFQ_INV(res, a, mun)    { ++_inv_call; (res) = (mun)-(a); (res)=(res)?(res):(mun); ++_inv_count; }

#define _GIVARO_GFQ_DIV(res, a, b, mun) {  ++_div_call; \
if ( (a)==0 ) { (res)=0; } else { (res) = (((res)=(a)-(b))>0)?(res):(res)+(mun); ++_div_count; } }



#define _GIVARO_GFQ_SQ(res, a, mun) { ++_mul_call; if ( (a)==0) (res) = 0; else \
                                  { (res) = ( (a) << 1) - (mun); \
                                    (res) = ((res)>0)?(res):(res)+ (mun); ++_mul_count; } }

// plun -> 1+^c - (q-1) !!!
// Warning : valid iff b != c
#define _GIVARO_GFQ_SQADD(c,a,b,mun,plun) { ++_mul_call; ++_add_call; \
if ((a)==0) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((  (c)=((a) << 1) - (mun)        )>0)?(c):(c) + (mun); \
++_mul_count; } else { \
       (c) = ((    (c) = ((a) << 1)-(b)-(mun)             )<0)?(c)+(mun):(c); \
       if (  (c) = (plun)[(UTT)(((c)>0)?(c):(c)+(mun))]     ) { \
                (c) = ((    (c) = (c)+(b)         )>0)?(c):(c)+(mun); } \
++_mul_count; ++_add_count;      }\
}

// Warning : valid iff b != c
#define _GIVARO_GFQ_MULADD(c,a1,a2,b,mun,plun) { ++_mul_call; ++_add_call; \
if (((a1)==0) || ((a2)==0)) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((    (c)=(a1)+(a2) - (mun)       )>0)?(c):(c) + (mun); \
++_mul_count; } else { \
       (c) = ((    (c) = (a1)+(a2)-(b)-(mun)        )<0)?(c)+(mun):(c); \
       if ( (c) = (plun)[(UTT)( ((c)>0)?(c):(c)+(mun)   )]  ) { \
                (c) = ((    (c) = (c)+(b)        )>0)?(c):(c)+(mun); }\
++_mul_count; ++_add_count;      }\
}

// Warning : valid iff b != c
#define _GIVARO_GFQ_MULSUB(c,a1,a2,b,mo,mun,plun) { ++_mul_call; ++_sub_call; \
if (((a1)==0) || ((a2)==0)) { (c)=(b); \
} else if ((b)==0) { \
       (c) = ((    (c)=(a1)+(a2) - (mo) -(mun)       )>0)?(c):(c) + (mun); \
       (c) = (c)>0?(c):(c) + (mun); \
++_mul_count; ++_neg_count;      } else { \
       (c) = ((    (c) = (a1)+(a2)-(b)-(mun) - (mo)       )<0)?(c)+(mun):(c); \
       (c) = (c)<0?(c)+(mun):(c); \
       if ( (c) = (plun)[(UTT)( ((c)>0)?(c):(c)+(mun)   )]  ) { \
                (c) = ((    (c) = (c)+(b)        )>0)?(c):(c)+(mun); }\
++_mul_count; ++_sub_count;      }\
}


#endif



template<class TT> 
inline GFqDom<TT>::Residu_t GFqDom<TT>::residu( ) const
{ return _q; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::cardinality( ) const
{ return _q; }
template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::characteristic( ) const
{ return _characteristic; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::exponent( ) const
{ return _exponent; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::size( ) const
{ return _q; }


 // ------------------------- Miscellaneous functions

template<class TT> 
inline bool GFqDom<TT>::isequal(const Rep a, const Rep b) const
  { return a == b ; }

template<class TT> 
inline bool GFqDom<TT>::isnequal(const Rep a, const Rep b) const
  { return a != b ; }

template<class TT> 
inline bool GFqDom<TT>::iszero(const Rep a) const
  { return a == GFqDom<TT>::zero ; }

template<class TT> 
inline bool GFqDom<TT>::isnzero(const Rep a) const
  { return a != GFqDom<TT>::zero ; }

template<class TT> 
inline bool GFqDom<TT>::isone(const Rep a) const
  { return a == GFqDom<TT>::one ; }

template<class TT> 
inline size_t GFqDom<TT>::length(const Rep ) const
  { return sizeof(TT) ;}

  // ----------- Usefull method :
template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::mul 
 (Rep& r, const Rep a, const Rep b) const 
{ _GIVARO_GFQ_MUL(r,a,b, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::mulin 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_MUL(r,r,a, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::div 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
{ _GIVARO_GFQ_DIV(r, a, b, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::divin 
(GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_DIV(r, r, a, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::add 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
{ _GIVARO_GFQ_ADD(r, a, b, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::addin 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_ADD(r, r, a, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::sub 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
{ _GIVARO_GFQ_SUB(r, a, b, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::subin 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_AUTOSUB(r, a, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::neg 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_NEG(r, a, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::negin 
 (GFqDom<TT>::Rep& r) const 
{ _GIVARO_GFQ_NEG(r, r, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::inv 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
{ _GIVARO_GFQ_INV(r, a, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::invin 
 (GFqDom<TT>::Rep& r) const 
{ _GIVARO_GFQ_INV(r, r, GFqDom<TT>::_qm1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::axpy 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b, const GFqDom<TT>::Rep c) 
 const 
{ _GIVARO_GFQ_MULADD(r,a,b,c, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::axpyin 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
{ 
  Rep tmp = r; 
  _GIVARO_GFQ_MULADD(r,a,b,tmp, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; 
return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::axmyin 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
{ 
//   Rep tmp = r; 
//   _GIVARO_GFQ_MULSUB(r,a,b,tmp, _qm1o2, _qm1, _plus1) ; 
   Rep tmp; _GIVARO_GFQ_MUL(tmp,a,b, _qm1) ;  
   _GIVARO_GFQ_AUTOSUB(r,tmp, _qm1o2, _qm1, _plus1) ;
return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::axmy 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b, const GFqDom<TT>::Rep c) 
 const 
{ 
  _GIVARO_GFQ_MUL(r,a,b, GFqDom<TT>::_qm1) ; 
  _GIVARO_GFQ_AUTOSUB(r,c, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; 
return r; }

template<class TT> 
inline GFqDom<TT>::Rep&  GFqDom<TT>::amxy 
 (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b, const GFqDom<TT>::Rep c) 
 const 
{ 
  _GIVARO_GFQ_MUL(r,a,b, GFqDom<TT>::_qm1) ; 
  _GIVARO_GFQ_SUB(r,c,r, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ; 
return r; }


 // -- inline array operations between ZpzDom<Log16>::Rep
template<class TT> 
inline void GFqDom<TT>::mul 
 (const size_t sz, Array r, constArray a, constArray b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MUL(r[i],a[i],b[i], GFqDom<TT>::_qm1) ; 
  }
}

template<class TT> 
inline void GFqDom<TT>::mul 
 (const size_t sz, Array r, constArray a, Rep b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MUL(r[i],a[i],b, GFqDom<TT>::_qm1) ; 
  }
}

template<class TT> 
inline void GFqDom<TT>::div 
 (const size_t sz, Array r, constArray a, constArray b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_DIV(r[i],a[i],b[i], GFqDom<TT>::_qm1) ; 
  }
}

template<class TT> 
inline void GFqDom<TT>::div 
 (const size_t sz, Array r, constArray a, Rep b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_DIV(r[i],a[i],b, GFqDom<TT>::_qm1) ; 
  }
}

template<class TT> 
inline void GFqDom<TT>::add 
 (const size_t sz, Array r, constArray a, constArray b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_ADD(r[i], a[i], b[i], GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::add 
 (const size_t sz, Array r, constArray a, Rep b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_ADD(r[i], a[i], b, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::sub 
 (const size_t sz, Array r, constArray a, constArray b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_SUB(r[i], a[i], b[i], GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::sub 
 (const size_t sz, Array r, constArray a, Rep b) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_SUB(r[i], a[i], b, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::neg 
 (const size_t sz, Array r, constArray a) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_NEG(r[i], a[i], GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::inv 
 (const size_t sz, Array r, constArray a) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_INV(r[i], a[i], GFqDom<TT>::_qm1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axpy 
(const size_t sz, Array r, Rep a, constArray x, constArray y) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MULADD(r[i], a, x[i], y[i], GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axpyin 
 (const size_t sz, Array r, Rep a, constArray x) const 
{
  Rep tmp;
  for (register size_t i=sz-1; i!=0; --i) {
    tmp = r[i];
    _GIVARO_GFQ_MULADD(r[i], a, x[i], tmp, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axpy 
 (const size_t sz, Array r, Rep a, constArray x, Rep y) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MULADD(r[i], a, x[i], y, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axmy 
 (const size_t sz, Array r, Rep a, constArray x, constArray y) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MUL(r[i], a, x[i], GFqDom<TT>::_qm1) ;
    _GIVARO_GFQ_AUTOSUB(r[i], y[i], GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axmy 
 (const size_t sz, Array r, Rep a, constArray x, Rep y) const 
{
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MUL(r[i], a, x[i], GFqDom<TT>::_qm1) ;
    _GIVARO_GFQ_AUTOSUB(r[i], y, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

template<class TT> 
inline void GFqDom<TT>::axmyin 
 (const size_t sz, Array r, Rep a, constArray x) const 
{
  Rep tmp;
  for (register size_t i=sz-1; i!=0; --i) {
    _GIVARO_GFQ_MUL(tmp, a, x[i], GFqDom<TT>::_qm1) ;
    _GIVARO_GFQ_AUTOSUB(r[i], tmp, GFqDom<TT>::_qm1o2, GFqDom<TT>::_qm1, GFqDom<TT>::_plus1) ;
  }
}

  // ------------------------------------
  // Input - Output  of the Domain 
  // 
template<class TT> 
inline istream& GFqDom<TT>::read (istream& s) {
    char ch;
    s >> ws >> ch;
    if (ch != '(')
        cerr << "GFqDom::read: syntax error: no '('" << endl;
    UTT p;
    s >> p;
    s >> ws >> ch;
    if (ch == ')') 
        *this = GFqDom<TT>(p,UTT(1));
    else {
        if (ch != '^')
            cerr << "GFqDom::read: syntax error: no '^'" << endl;
        UTT k;
        s >> ws >> k;
        if (ch != ')')
            cerr << "GFqDom::read: syntax error: no ')'" << endl;

        *this = GFqDom<TT>(p,k);
    }
    return s;
}   

template<class TT> 
inline ostream& GFqDom<TT>::write (ostream& o) const {
    return o << '(' <<  GFqDom<TT>::_characteristic << '^' << GFqDom<TT>::_exponent << ')' << endl;
}

  // ------------------------------------
  // Input - Output  of the elements
  // 
template<class TT> 
inline istream& GFqDom<TT>::read (istream& i, Rep& a) const {
    TT t;
    i >> t;
    read(a,t);
    return i;
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::read( Rep& r, const int residu ) const {
    return read (r, TT(residu));
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::read( Rep& r, const unsigned int residu ) const {
    return read (r, UTT(residu));
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r, const double residu ) const {
  double tr = residu ;
  if (tr <0) {
      // -a = b [p]
      // a = p-b [p]
    tr = -tr;
    if (tr >= (TT)_characteristic ) tr = (UTT)tr % _characteristic ;
    if (tr)
      return r = _pol2log[ _characteristic - (UTT)tr ];
    else
      return r = zero;
  } else {
    if (tr >= (TT)_characteristic ) tr = (UTT)tr % _characteristic ;
    return r = _pol2log[ tr ];
  }
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::read( Rep& r, const TT residu ) const {
  TT tr = residu ;
  if (tr <0) {
      // -a = b [p]
      // a = p-b [p]
    tr = -tr;
    if (tr >= (TT)_characteristic ) tr = (UTT)tr % _characteristic ;
    if (tr)
      return r = _pol2log[ _characteristic - (UTT)tr ];
    else
      return r = zero;
  } else {
    if (tr >= (TT)_characteristic ) tr = (UTT)tr % _characteristic ;
    return r = _pol2log[ tr ];
  }
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::read( Rep& r, const Integer residu ) const 
{
  UTT tr;
  if (residu <0) {
      // -a = b [p]
      // a = p-b [p]
    if ( residu <= (Integer)(-_characteristic) ) tr =  (-residu) % (UTT)_characteristic ;
    else tr = Integer2long(-residu);
    if (tr)
      return r = _pol2log[ _characteristic - (UTT)tr ];
    else
      return r = zero;
  } else {
    if (residu >= (Integer)_characteristic ) tr =  residu % (UTT)_characteristic ;
    else tr = Integer2long(residu);
    return r = _pol2log[ tr ];
  }
}

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::read( Rep& r, const UTT residu ) const 
{
    UTT tr = residu ;
    if (tr >= _characteristic ) tr = tr % _characteristic ;
    return r = _pol2log[ tr ];
}

template<class TT> 
inline ostream& GFqDom<TT>::write (ostream& o, const Rep a) const {
  return o << _log2pol[ (UTT)a] ;
}

template<class TT> 
inline double& GFqDom<TT>::convert (double& r, const Rep a) const 
{
  return r = (double)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline TT& GFqDom<TT>::write (TT& r, const Rep a) const 
{
  return r = (TT)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline GFqDom<TT>::UTT& GFqDom<TT>::write (UTT& r, const Rep a) const 
{
  return r = (UTT)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline int& GFqDom<TT>::write (int& r, const Rep a) const 
{
  return r = (int)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline unsigned int& GFqDom<TT>::write (unsigned int& r, const Rep a) const 
{
  return r = (unsigned int)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline TT GFqDom<TT>::write (const Rep a) const 
{
  return (TT)_log2pol[ (UTT)a] ;
}

template<class TT> 
inline Integer& GFqDom<TT>::write (Integer& r, const Rep a) const 
{
  return r = (Integer)_log2pol[ (UTT)a] ;
}


// ---------
// -- misc operations
// ---------
template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r) const { return r = zero; }

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r, const TT a) const { return read (r, a); }

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r, const UTT a) const { return read (r, a); }

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r, const Integer a) const { return read (r, a); }

//template<class TT> 
//inline GFqDom<TT>::Rep& GFqDom<TT>::init( Rep& r, const Rep a) const { return r = Rep(a); }


// template<class TT> 
// inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const TT a) const { return read (r, a); }

// template<class TT> 
// inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const UTT a) const { return read (r, a); }

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const Integer a) const { return read (r, a); }

template<class TT> 
inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const Rep a) const { return r = a; }


template<class TT> inline void GFqDom<TT>::assign( const size_t sz, Array r, constArray a ) const {
    TT tr;
//    for (register size_t i=sz-1; i!=0; --i) {
    for (register size_t i=sz; i--;) { 
        tr = a[i] ;
        if (tr <0) {
                // -a = b [p]
                // a = p-b [p]
            tr = -tr;
            if (tr >=_characteristic ) tr = tr % _characteristic ;
            if (tr)
                r[i] = _pol2log[ _characteristic - tr ];
            else
                r[i] = 0;
        } else {
            if (tr >=_characteristic ) tr = tr % _characteristic ;
            r[i] = _pol2log[ tr ];
        }
    }
}

template<class TT> 
inline  GFqDom<TT>::Rep& GFqDom<TT>::dotprod 
 ( Rep& r, const size_t sz, constArray a, constArray b ) const 
{
  if (sz) {
    _GIVARO_GFQ_MUL(r,a[0],b[0],_qm1);
    Rep tmp;
    for( register int i= sz-1; i>0; --i) {
      _GIVARO_GFQ_MUL(tmp,a[i],b[i],_qm1);
      _GIVARO_GFQ_ADD(r,r,tmp,_qm1,_plus1);
    }
    return r;
  } else
    return r = zero;
}

    
   // ----- random generators
template<class TT> template<class RandIter> inline GFqDom<TT>::Rep& GFqDom<TT>::nonzerorandom(RandIter& g, Rep& a) const {
//     do 
//         a = Rep( (UTT)(lrand48()) % _q);
//     while (iszero(a));
//     a = (a<0?a+_q:a);
//     return a;
    a = Rep( ((UTT)(g()) % (_q-1)) + 1);
    return a = (a<0?a+_q:a);
    
}

template<class TT> template<class RandIter> inline GFqDom<TT>::Rep& GFqDom<TT>::random(RandIter& g, Rep& a) const {
    a = Rep( (UTT)(g()) % _q);
    return a = (a<0?a+_q:a);
}
    
template<class TT> template<class RandIter> 
inline GFqDom<TT>::Rep& GFqDom<TT>::random(RandIter& g, Rep& r, long s) const { 
    return random(g,r);
}

    
template<class TT> template<class RandIter> 
inline GFqDom<TT>::Rep& GFqDom<TT>::random(RandIter& g, Rep& r, const Rep& b) const { 
    return random(g,r);
}

template<class TT> template<class RandIter> 
inline GFqDom<TT>::Rep& GFqDom<TT>::nonzerorandom(RandIter& g, Rep& r, long s) const { 
    return nonzerorandom(g,r);
}

template<class TT> template<class RandIter> 
inline GFqDom<TT>::Rep& GFqDom<TT>::nonzerorandom(RandIter& g, Rep& r, const Rep& b) const {
    return nonzerorandom(g,r);
}



// ==========================================================================
// file: givgfq.C
// Time-stamp: <20 Dec 99 11:18:32 Jean-Guillaume.Dumas@imag.fr>
// ==========================================================================
// #include "givgfq.h"
#ifndef __GIVARO_GFQ_C__
#define __GIVARO_GFQ_C__

#include <givinteger.h>
#include <givintnumtheo.h>
#include <givpower.h>
#include <givpoly1padic.h>
#include <givpoly1factor.h>

#include <vector.h>

template<class TT>
inline GFqDom<TT>::GFqDom(const UTT P, const UTT e = 1) 
        // Precondition P prime
        :  zero(0)
    , one (mod_power(P,e) - 1  )
    , _characteristic(P)
    , _exponent(e) 
    , _q( one + 1 )
    , _qm1 ( one )
    , _qm1o2(  (P==2)?  (one)  :  (_q >> 1) )   // 1 == -1 in GF(2^k)
    , _log2pol( _q )
    , _pol2log( _q )
    , _plus1( _q )
{
        // 1 is represented by q-1, zero by 0
    _log2pol[0] = zero;

    if (e == 1) {
        IntNumTheoDom<> NTD;
        IntNumTheoDom<>::Rep IP(P), pr;
        UTT seed = (UTT) ( NTD.Integer2long( NTD.lowest_prim_root(pr, IP) ) );
        UTT accu = 1;
        for(UTT i=1; i<P; i++) {
            accu = (accu * seed) % P;
            _log2pol[i] = accu;
        }      
    } else {
            // Fisrt compute an irreductible polynomial F over Z/pZ of degree e
            // Then a primitive root G (i.e. a generator of GF(q))
        GFqDom<TT> Zp(P);
        typedef Poly1FactorDom< GFqDom<TT>, Dense > PolDom;
//         typedef CyclotomicTable<  GFqDom<TT>, Dense > PolDom;
//         PolDom Pdom( Zp, e );
        PolDom Pdom( Zp );
        typename PolDom::element F, G, H;

            // F is irreducible of degree e over Zp
            // G is a primitive polynomial for F
//         Pdom.random_prim_root(F,G, Degree(e));

            // F is an irreducible factor of the 
            // (p^e-1) th cyclotomic polynomial
            // G is a primitive polynomial for F : X
//         Pdom.getcyclo(F);
//         Pdom.init(G, Degree(1), Zp.one);

            // F is irreducible of degree e over Zp
            // with X as a primitive polynomial
#ifndef GIVARO_RANDOM_IRREDUCTIBLE_PRIMITIVE_ROOT
        Pdom.ixe_irreducible(F, Degree(e));
        Pdom.init(G, Degree(1), Zp.one);
#else
        Pdom.random_irreducible(F, Degree(e));
        Pdom.give_random_prim_root(G,F);
#endif     
        Pdom.assign(H, G);

        typedef Poly1PadicDom< GFqDom<TT>, Dense > PadicDom;
        PadicDom PAD(Pdom);
         
        PAD.eval(_log2pol[1], H);
        
        for (UTT i = 2; i < _qm1; ++i) {
            Pdom.mulin(H, G);
            Pdom.modin(H, F); 
            PAD.eval(_log2pol[i], H);
        }
        
        _log2pol[_qm1] = 1;

    }

    _log2pol[0] = 0;

        // pol2log[ j ] = i such that log2pol[i] = j
    for (UTT i = 0; i < _q; ++i) 
        _pol2log[ _log2pol[i] ] = i;

        // plus1[i] = k such that G^i + 1 = G^k
        // WARNING : in the plus1 table, we now pre-substract (_q - 1)
    _plus1[0] = 0;

    UTT a,b,r;
    for (UTT i = 1; i < _q; ++i) {
        a = _log2pol[i];
        r = a % P;
        if (r == (P - 1))
            b = a - r;
        else
            b = a + 1;
            // WARNING : in the plus1 table we pre-substract (_q - 1)
        _plus1[i] = _pol2log[b] - _qm1; 
    }
        // -1 + 1 == 0
   _plus1[_qm1o2] = 0; 
}


template<class TT> inline void GFqDom<TT>::Init() {} 

template<class TT> inline void GFqDom<TT>::End() {}


#ifdef __GIVARO_COUNT__
template<class TT> long long GFqDom<TT>::_mul_count = 0;
template<class TT> long long GFqDom<TT>::_add_count = 0;
template<class TT> long long GFqDom<TT>::_div_count = 0;
template<class TT> long long GFqDom<TT>::_sub_count = 0;
template<class TT> long long GFqDom<TT>::_neg_count = 0;
template<class TT> long long GFqDom<TT>::_inv_count = 0;
template<class TT> long long GFqDom<TT>::_mul_call = 0;
template<class TT> long long GFqDom<TT>::_add_call = 0;
template<class TT> long long GFqDom<TT>::_div_call = 0;
template<class TT> long long GFqDom<TT>::_sub_call = 0;
template<class TT> long long GFqDom<TT>::_neg_call = 0;
template<class TT> long long GFqDom<TT>::_inv_call = 0;
#endif

#endif

    
