#ifndef __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
#define __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
// ======================================================================= //
// Copyright (C) LinBox project 1999
// Givaro / Athapascan-1
// Smith form
// Time-stamp: <25 Feb 02 14:10:53 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <list.h>

#include "givintfactor.h"
#include "givintrns.h"
#include "givgfq.h"


#include "LinBox/lin_dom_gauss.C"
#include "LinBox/lin_pp_gauss.C"
#include "LinBox/lin_density.h"
#include "LinBox/lin_invariants.h"
#include "LinBox/lin_spv_bb.h"
#include "LinBox/lin_erase_bb.h"
#include "LinBox/lin_wrap_c++.h"


typedef SparseBlackBoxDom< OperatorWrapper<long> > SPBBlong;
typedef SparseBlackBoxEraseDom< GFqDomain > SPBBE ;
    
void one_gauss(unsigned long& rank, char * Matrix_File_Name, unsigned long MOD) {
    
    GFqDomain F(MOD, 1);
    GaussDom< GFqDomain > GF( F, GlobalCommentator );    

    SPBBE TheMat(F);
    unsigned long ni, nj, ne, addrank; TheMat.read(ni,nj,Matrix_File_Name);
    if (ni > nj) {
        addrank = TheMat.read_erase(Matrix_File_Name);
        char *UT, * File_Name;
        File_Name = tempnam("/tmp","bbx_"); // "/tmp/bbx_XXXXXX"; mkstemp(File_Name);
        TheMat.write(File_Name);
        TheMat.read_transpose(File_Name);
        UT = new char[5+strlen(File_Name)];
        sprintf(UT,"\\rm %s", File_Name);
        system(UT);
        delete [] UT; delete [] File_Name;
    } else
        addrank = TheMat.read_erase(Matrix_File_Name);

    { GF.gauss_rankin(rank,TheMat,Density()); }
    if (addrank > 0) rank+=addrank;
}

void one_permuts(Invariants<unsigned long>& c_rows, unsigned long& rank, char * Matrix_File_Name, unsigned long MOD) {
    
    GFqDomain F(MOD, 1);
    GaussDom< GFqDomain > GF( F, GlobalCommentator );    

    SPBBE TheMat(F);
    unsigned long ni, nj, ne; 
    char *UT, * File_Name;
    TheMat.read_transpose(Matrix_File_Name);

    GF.gauss_rankin(rank,TheMat,Density(), 1);

    for(unsigned long i = 0; i< TheMat.size(); ++i)
        if (TheMat[i].size() > 0) c_rows.PushInvariant(i, 1);
}

#define GIVARO_MACHINT signed long
#define GIVARO_MACHIND unsigned long
void one_pp_gauss(unsigned long& rank, char * Matrix_File_Name, long PRIMEPOWER, long PRIME) {
        OperatorWrapper<GIVARO_MACHINT> F;
        SparseBlackBoxDom< OperatorWrapper<GIVARO_MACHINT>, GIVARO_MACHIND >  MF( F );
        SparseBlackBoxDom< OperatorWrapper<GIVARO_MACHINT>, GIVARO_MACHIND >::element TheMat;
        GIVARO_MACHIND ni, nj, nz;
        MF.read(ni, nj, Matrix_File_Name);
        if (ni > nj)
            MF.read_transpose(TheMat, ni, nj, nz, Matrix_File_Name);
        else
            MF.read(TheMat, ni, nj, nz, Matrix_File_Name);
	
	// MF.write( "Toto", TheMat );

        prime_power_rankin(PRIMEPOWER, PRIME, rank, TheMat, ni, nj, Density());
}


void gauss_Integer_rank(ulong& rank, Integer& p1, char * File_Name, Integer IN, double& wrong, double error) {
    IntPrimeDom IPD;
    IPD.prevprime(p1, 1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/20);
    while ( IPD.iszero(IN % p1)) IPD.prevprimein(p1);
    one_gauss(rank, File_Name, (long)p1);
}




void gauss_rank_N_torsion(Invariants<ulong>& Inv, char * File_Name, const Integer& IN, bool withintegerrank = 0) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    unsigned long rank;
    Integer valence;
    ulong nbfactor;
    list<Integer> Lm;
    list<ulong> Le;
    
    IntFactorDom<> IFD;
    nbfactor = IFD.set(Lm,Le,IN);

    list<Integer>::iterator im=Lm.begin();
    list<ulong>::iterator ie=Le.begin();

    OperatorWrapper<long> LongDom;
    SPBBlong TheMat( LongDom );
    unsigned long ni, nj; TheMat.read(ni,nj,File_Name);

    unsigned long * r = new unsigned long[nbfactor];
    bool * t = new bool[nbfactor];
    unsigned long ir;
    unsigned long k, dimr = GIVMIN(ni,nj);
    unsigned long minr = dimr, maxr = 0;
    
    char * UT = new char[170]; 


    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
        if (*ie > 0) {
            pr *= *im;
            if (*im < Integer(MAXGFQSIZE)) {
                t[k] = true;
                one_gauss( r[k], File_Name, (long)(*im));
                minr = GIVMIN(minr,r[k]);
                maxr = GIVMAX(maxr,r[k]);
                sprintf(UT,"rank mod %ld : %ld", (long)(*im), r[k]);
            } else {
                t[k] = false;
                string s; Integer2string(s, *im);
                sprintf(UT,"WARNING, %s is TOO BIG. Supposed full rank, WARNING.", s.c_str());
            }
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
           
        }
    
    if (withintegerrank) {
        Integer p(2);
        while ((iszero(pr % p)) || (! IFD.isprime(p)))
            IFD.nextprimein(p);
        one_gauss( maxr, File_Name, (long)p);
        sprintf(UT,"Integer rank (mod %ld) : %ld", (long)(p), maxr);
        activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        ir = maxr;
    } else {
        ir = dimr;
    }
    
    
    if (ir > minr) {
        unsigned long ssize = ir-minr;
        ulong * Smith = new ulong[ssize];
        unsigned long ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        unsigned long l=0,l1=0, count_ie;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
            if (t[k]) {
                if (*ie > 0) {
                    count_ie = 0;
                    l = r[k]; 
                    ulong p = (long)(*im);
                    ulong pp = p;
//                 while (l < ir) {
                    while ( (l<ir) || (!withintegerrank) ) {
                        for(unsigned long lk=ir-l;lk;--lk)
                            Smith[ssize-lk] *= p;
                        pp *= p;
                        if (++count_ie < *ie) {
                            one_pp_gauss(l, File_Name, pp,p);
                            maxr = GIVMAX(maxr, l);
                        } else
                            break;
                        sprintf(UT,"pp rank mod %ld : %ld", pp, l);
                        activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
                    }
                }
            }

        Inv.PushInvariant(1,minr);
        unsigned long lj=0;
        ssize = GIVMIN(maxr, dimr) - minr;
        for(ll=0;ll<ssize;) {
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
//         delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
    }
    
        
//     delete [] r; delete [] t; delete [] UT;
    stopActivity(GlobalCommentator, "Done.", LVL_BLABLA, INTERNAL_DESCRIPTION ); 
};




#endif 
