// ======================================================================= //
// Copyright (C)  1999, Linbox project
// Givaro / Athapascan-1
// Valence computation
// Time-stamp: <04 Oct 01 18:30:44 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //
#ifndef __LINBOX_VALENCE_C__
#define __LINBOX_VALENCE_C__

#include <math.h>
#define NUMBEROFPRIMES 3030.0
#define MAXAVAILABLEPRIMES 6103
#define MAXPRIMESIZE 65536
#define MINLOWERSIZE 5000
#define valBOUND 5

#include "givintprime.h"
#include "givintrns.h"
#include "givgfq.h"
typedef GFqDom<long>               GFqDomain;
typedef GFqDomain::element       Residu;

#include "LinBox/lin_rand.h"                      // Random Iterator
#include "LinBox/lin_spv_bb.h"                    // BB Wrapper for sparse vectors
#include "LinBox/lin_symmetrize_bbit.h"           // BB iterator
#include "LinBox/lin_massey.C"                // massey reccuring sequence solver

typedef SparseBlackBoxDom< GFqDomain > SPBB ;
    // Compute the rank of B^T B
    // with B == D1 A^T D2^2 A D1
typedef BB_Symmetrize_Container< SPBB > SzCBB;



void one_valence(char * File_Name, Random& generator, unsigned long& v, unsigned long& p1, unsigned long& rk1) {
    IntPrimeDom IPD; Integer ip1;
    do {
        ip1 = p1;
        IPD.prevprimein(ip1);
        p1 = (long)ip1;
    } while (! IPD.isprime(p1));
    
    GFqDomain F(p1, 1);
    GFqDomain::element v1;
    SPBB MF( F );
    MF.read(File_Name);
    {
        SzCBB TF( &MF, generator); 
        MasseyDom< SzCBB >  WD(GlobalCommentator, &TF );
        WD.valence(v1, rk1);
    }
    v = F.convert(v1);
}



void seq_Integer_valence(char * File_Name, Integer& Valence, unsigned long & rank, double bound, double& wrong, double error = 0) {

    startActivity(GlobalCommentator, "Int Vale", "Integer Valence", LVL_NORMAL, INTERNAL_DESCRIPTION ); 

    double ulogMM = 1/log(MAXPRIMESIZE - MINLOWERSIZE);

    
    unsigned long rk1, rk2;
    unsigned long valence, v;
    unsigned long r;


    vector<Integer> Lr, Lm;

    Random generator;
    
    unsigned long p1 = MAXPRIMESIZE - generator() % (ulong)MINLOWERSIZE; 
    one_valence(File_Name, generator, valence, p1, rank);
    char * UT = new char[500]; sprintf(UT,"Min Poly Degree : %ld", rank);
    activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT);
    Lr.push_back(Integer(valence));
    Lm.push_back(Integer(p1));
   
    double maxsize = (double)rank * log(bound) * ulogMM;
    double bad = (double)(rank+1)*double(rank+1) * log(bound) / log(MINLOWERSIZE) / 2.0;
    if ((error > 0) && (error < 1)) {
	double logbad = GIVMAX( 1, log( bad / 3030.0 ) ); 
	double maxerror = log ( error ) / logbad;
        if (maxerror > maxsize) { 
            activityReport(GlobalCommentator, "Taking high probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
            maxsize = maxerror;
        }
    }
    unsigned long nbloops = (unsigned long)ceil(maxsize);
    if (nbloops > MAXAVAILABLEPRIMES) {
        Valence = Integer(1);
        activityReport(GlobalCommentator, "More than 6103 primes needed. Wanted Probability too high.", LVL_IMP, INTERNAL_ERROR);
    } else {
	sprintf(UT,"%ld modular valences to be computed",nbloops);
	activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION);
    }
   
    Valence = Integer(valence);
    if (rank & 0x1) {
        if (Valence > 0)
            Valence -= p1;
    } else
        if (Valence < 0)
            Valence += p1;

    for(int l=nbloops-1;l>0;--l) {
        progressReport(GlobalCommentator, "modular valence", LVL_NORMAL, nbloops-l, nbloops);
        one_valence(File_Name, generator, v, p1, r);

        if (r != rank) { ++l; continue; }
        
        Lr.push_back(Integer(v));
        Lm.push_back(Integer(p1));
        IntRNSsystem<vector> RNs( Lm );
        IntRNSsystem<vector>::element V0;
        RNs.RnsToRing(V0, Lr);
        if (rank & 0x1) {
            if (V0 > 0)
                V0 -= RNs.product();
        } else
            if (V0 < 0)
                V0 += RNs.product();
//            if (V0 == Valence) if (++diff_bound > valBOUND) break;
        Valence = V0;
    }
    
    wrong = pow( bad/NUMBEROFPRIMES, (double)nbloops);
    string s; Integer2string(s, Valence);
    UT= new char[500]; sprintf(UT,"Valence is %s, correct with probability at least 1 - %e", s.c_str(), wrong ); 
    stopActivity( GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION) ;
    delete [] UT;
}

#endif 
